from itertools import product
from typing import Optional, Tuple

import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr


def _arange_center(start, stop, size):
    """Generate evenly spaced values centered in each interval."""
    step = (stop - start) / size
    return np.arange(start + step / 2, stop + step / 2, step)


def _narray_indices(shape):
    """Generate a NumPy array containing the index values."""
    narray = np.empty(shape, dtype=object)
    for indices in product(*(range(s) for s in shape)):
        narray[indices] = str(indices)
    return narray


def _narray_random(shape, seed=None):
    """Generate a NumPy array with random values."""
    rs = np.random.default_rng(seed)
    return rs.random(shape)


def _darray_indices(shape, chunks=None):
    """Generate a Dask array containing the index values."""
    narray = _narray_indices(shape)
    darray = da.from_array(narray, chunks=chunks)
    return darray


def _darray_random(shape, chunks=None, seed=None):
    """Generate a Dask array with random values."""
    rs = da.random.RandomState(seed=seed)
    return rs.random(shape, chunks=chunks)
    # return da.random.random(shape, chunks=chunks)


def array(shape, chunks=None, fmt="drandom", seed=None):
    """Generate an array of the requested type and format."""
    if fmt == "nrandom":
        return _narray_random(shape, seed)
    elif fmt == "nindices":
        return _narray_indices(shape)
    elif fmt == "drandom":
        return _darray_random(shape, chunks, seed)
    elif fmt == "dindices":
        return _darray_indices(shape, chunks)
    else:
        raise ValueError(
            f"Format '{fmt}' not recognized. Use 'nrandom', 'nindices', 'drandom', or 'dindices'."
        )


def generate_grid(
    nvar: int,
    shape: Tuple[int, ...],
    chunks=None,
    fmt: str = "random",
    seed: Optional[int] = None,
) -> xr.Dataset:
    """Generate an xarray Dataset with coordinates and random variables."""

    ndim = len(shape)
    dims = [f"dim{i}" for i in range(ndim)]

    ds = xr.Dataset()

    # Create simple coordinates
    for i, size in enumerate(shape):
        ds[dims[i]] = np.arange(size)

    # Create random variables
    for n in range(nvar):
        data = array(shape, chunks, fmt, seed)
        ds[f"var{n}"] = (dims, data)

    return ds


def generate_spatio_temporal_grid(
    nvar: int,
    shape: Tuple[int, int, int],
    chunks=None,
    fmt: str = "random",
    seed: Optional[int] = None,
) -> xr.Dataset:
    """Generate a spatio-temporal xarray Dataset with random variables."""

    # Dimensions and coordinates
    ntime, nlon, nlat = shape
    coords = {
        "time": pd.date_range("2000-01-01", periods=ntime),
        "lon": _arange_center(-180, 180, nlon),
        "lat": _arange_center(-90, 90, nlat),
    }

    dims = ("time", "lon", "lat")

    # Create the Dataset
    ds = xr.Dataset(coords=coords)

    # Add random variables
    for n in range(nvar):
        data = array(shape, chunks, fmt, seed)
        ds[f"var{n}"] = (dims, data)

    return ds



import numpy as np
import xarray as xr

# def is_array_like(obj):
#     """Vérifie si un objet est de type list, np.ndarray ou xarray.DataArray."""
#     return isinstance(obj, (list, np.ndarray, xr.DataArray))

# def ensure_array_like(x):
#     """Convertit un scalaire en liste d’un élément si besoin."""
#     if is_array_like(x):
#         return x
#     return [x]


def ensure_numpy_array(x):
    """Converts an array-like object into a np.ndarray."""
    if isinstance(x, np.ndarray):
        return x
    elif isinstance(x, xr.DataArray):
        return x.values
    elif isinstance(x, list):
        return np.array(x)
    else:
        return np.array([x])



import numpy as np

EARTH_CIRCUMFERENCE_IN_DEG = 360
EARTH_CIRCUMFERENCE_IN_KM = 40_070


def km2deg(km):
    """Convert a distance from kilometers to degrees."""
    return (km * EARTH_CIRCUMFERENCE_IN_DEG) / EARTH_CIRCUMFERENCE_IN_KM


def deg2km(deg):
    """Convert a distance from degrees to kilometers."""
    return (deg * EARTH_CIRCUMFERENCE_IN_KM) / EARTH_CIRCUMFERENCE_IN_DEG


def nano2min(nano):
    """Convert time from nanoseconds to minutes."""
    return nano / np.timedelta64(1, "m")


def find_nearest_indices(coord_array: np.ndarray, obs_array: np.ndarray) -> np.ndarray:
    """Find nearest indexes of obs in coords array"""

    # Trouver les indices d'insertion
    idx = np.searchsorted(coord_array, obs_array)

    # Assurer que les indices sont dans les limites
    idx = np.clip(idx, 1, len(coord_array) - 1)

    # Comparer les valeurs avant et après l'index trouvé
    left = coord_array[idx - 1]
    right = coord_array[idx]

    # Trouver quel côté est le plus proche
    idx_nearest = idx - (obs_array - left < right - obs_array)

    return idx_nearest



from typing import Optional, Union

import numpy as np
import xarray as xr

from pytcube.utils.core import find_nearest_indices
from pytcube.utils.types import ensure_numpy_array

ArrayLike = Union[xr.DataArray, np.ndarray, list, int, float]


def func_array(values, **kwargs):
    """Generates an array of ranges centered around each value with a specified buffer."""
    return np.array(
        [
            np.arange(value - kwargs["buffer"], value + kwargs["buffer"] + 1)
            for value in values
        ]
    )


def func_in_domain(values, **kwargs):
    """Checks if values are within a specified range."""
    return np.array(
        [((value >= kwargs["min"]) and (value < kwargs["max"])) for value in values]
    )


@xr.register_dataset_accessor("point")
class PointAccessor:
    def __init__(self, xarray_obj: xr.Dataset):
        self._obj = xarray_obj

    def _format(self, prefix):
        vars_renamed = {dim: f"{prefix}{dim}" for dim in self._obj.data_vars}
        _obj = self._obj.rename_vars(vars_renamed)  # rename vars
        return _obj.assign_coords(
            {f"{prefix}{dim}": _obj[f"{prefix}{dim}"] for dim in self._obj.data_vars}
        )

    def _find_nearest_indices(
        self, grid: xr.Dataset, dim: str, grid_prefix: str, point_prefix
    ) -> xr.Dataset:
        """Finds the nearest indices in the grid for the given dimension."""
        coord_array = grid[f"{grid_prefix}{dim}"].values
        random_coords = self._obj[f"{point_prefix}{dim}"].values

        idx_nearest = find_nearest_indices(coord_array, random_coords)

        return xr.DataArray(
            idx_nearest.astype("int32"), dims=["obs"], coords={"obs": self._obj["obs"]}
        )

    def _in_domain(self, dataarray: xr.DataArray, _min: int, _max: int):
        """Checks if the values in a DataArray are within specified minimum and maximum bounds."""
        return xr.apply_ufunc(
            func_in_domain,
            dataarray,
            input_core_dims=[[]],
            output_core_dims=[[]],
            output_dtypes=[bool],
            dask="parallelized",
            kwargs={"min": _min, "max": _max},
        )

    def _array(
        self, dataarray: xr.DataArray, output_core_dims: str, buffer: dict[str, int]
    ):
        """Generates an array of ranges centered around each value with a specified buffer."""
        return xr.apply_ufunc(
            func_array,
            dataarray,
            input_core_dims=[[]],
            output_core_dims=[[output_core_dims]],
            output_dtypes=[int],
            dask_gufunc_kwargs=dict(output_sizes={output_core_dims: 2 * buffer + 1}),
            dask="allowed",
            kwargs={"buffer": buffer},
        ).astype("int32")


def set_points(obs: Optional[ArrayLike] = None, **indexers: ArrayLike) -> xr.Dataset:
    """Create a point Dataset conforming to PytCube specifications.

    The Dataset will have:
        * A coordinate 'obs'
        * Data variables corresponding to the dimensions to index in the datacube
        * No extra data variables beyond the indexed dimensions
    """

    # Convert all indexers to numpy arrays
    indexers = {k: ensure_numpy_array(v) for k, v in indexers.items()}
    if obs is not None:
        obs = ensure_numpy_array(obs)

    # Verify that all indexers have the same length
    lengths = [len(v) for v in indexers.values()]
    if obs is not None:
        lengths.append(len(obs))

    if len(set(lengths)) > 1:
        raise ValueError(f"The lengths of lists/arrays are not identical: {lengths}")

    # Create obs if it is not provided
    n = lengths[0] if lengths else 0
    if obs is None:
        obs = np.arange(n)

    return xr.Dataset(
        coords={"obs": obs}, data_vars={k: ("obs", v) for k, v in indexers.items()}
    )



from typing import Any, Dict, List, Optional

import xarray as xr
from fast_vindex import patched_vindex

from pytcube.core.point import set_points


@xr.register_dataset_accessor("pytcube")
class PytCubeAccessor:
    def __init__(self, xarray_obj: xr.Dataset):
        """Initialize the PytCube accessor with an xarray Dataset."""
        self._obj = xarray_obj

    def _colocation(
        self,
        point: xr.Dataset,
        search_window: Optional[Dict[str, int]] = None,
        cyclic_dims: List[str] = [],
        grid_prefix: str = "grid_",
        point_prefix: str = "point_",
        fast_vindex: bool = True,
    ):
        """Internal method .colocation()."""
        grid, point = self._obj, point

        grid_dims = grid.dims
        point_dims = (
            point.data_vars
        )  # Permet de connaitre les dimensions de points (avant que soit ajouter les idim, idim_array, ...)

        # On part du postulat que point_dims ne contient que les dimensions à extraire (vérifié dans .colocation())
        # Ajouter des tests pour vérifier que grid et points respecte les specs

        # Default search window: nearest point only
        if search_window is None:
            search_window = {dim: 0 for dim in grid_dims}

        # Check that all point dimensions exist in the grid
        missing_keys = [k for k in point_dims if k not in grid_dims]
        if missing_keys:
            raise ValueError(f"Missing keys in grid dataset: {missing_keys}")

        # Format the point and grid datasets
        point = point.point._format(prefix=point_prefix)
        grid = grid.grid._format(prefix=grid_prefix)

        # Find nearest indices for each point dimension
        for dim in point_dims:
            point[f"{point_prefix}i{dim}"] = point.point._find_nearest_indices(
                grid, dim, grid_prefix, point_prefix
            )

        # Check if the mini-cube is within domain
        for dim in point_dims:
            _min = search_window[dim]
            _max = grid.sizes[f"{grid_prefix}{dim}"] - search_window[dim]

            point[f"{point_prefix}i{dim}_in_domain"] = point.point._in_domain(
                point[f"{point_prefix}i{dim}"], _min, _max
            )

        # Filter points outside the domain
        non_cyclic_dims = [dim for dim in point_dims if dim not in cyclic_dims]
        condition = point[f"{point_prefix}i{non_cyclic_dims[0]}_in_domain"]  # == True
        for dim in non_cyclic_dims[1:]:
            condition &= point[f"{point_prefix}i{dim}_in_domain"]  # == True

        point = point.where(condition, drop=True)

        # Create the i_arrays for indexing
        for dim in point_dims:
            point[f"{point_prefix}i{dim}_array"] = point.point._array(
                point[f"{point_prefix}i{dim}"], f"{dim}_step", search_window[dim]
            )

        # Apply modulo for cyclic dimensions
        for dim in cyclic_dims:

            def modulo(x):
                return x % grid.sizes[f"{grid_prefix}{dim}"]

            # modulo = lambda x: x % grid.sizes[f"{grid_prefix}{dim}"]
            point[f"{point_prefix}i{dim}_array"] = modulo(
                point[f"{point_prefix}i{dim}_array"].astype("int32")
            )

        # Build the indexers dictionary
        indexers = {
            f"{grid_prefix}{dim}": point[f"{point_prefix}i{dim}_array"].astype("int32")
            for dim in point_dims
        }

        # Select data using either fast_vindex or standard indexing
        if fast_vindex:
            with patched_vindex():
                result = grid.isel(**indexers)
        else:
            result = grid.isel(**indexers)

        return result

    def colocation(
        self,
        obs: Optional[Any] = None,
        search_window: Optional[Dict[str, int]] = None,
        cyclic_dims: List[str] = [],
        grid_prefix: str = "grid_",
        point_prefix: str = "point_",
        fast_vindex: bool = True,
        **indexers_kwargs: Any,
    ):
        """
        Select grid data based on provided indices or points.

        Parameters
        ----------
        obs : array-like or None
            Observation indices.
        search_window : dict, optional
            Window size to extract around each point.
        cyclic_dims : list of str
            Dimensions that are cyclic.
        grid_prefix : str
            Prefix used for grid coordinates.
        point_prefix : str
            Prefix used for point coordinates.
        fast_vindex : bool
            Whether to use fast_vindex for indexing.
        **indexers_kwargs : array-like
            Coordinates or indexers to select along each dimension.

        Returns
        -------
        xr.Dataset
            Subset of the grid corresponding to the selected points.
        """

        indexers = indexers_kwargs

        # Check that all indexers exist in the dataset
        for dim in indexers:
            if dim not in self._obj.dims:
                raise ValueError(
                    f"Dimension '{dim}' non présente dans le dataset {list(self._obj.dims)}"
                )

        # Create a points Dataset
        point = set_points(obs=obs, **indexers)

        return self._colocation(
            point,
            search_window=search_window,
            cyclic_dims=cyclic_dims,
            grid_prefix=grid_prefix,
            point_prefix=point_prefix,
            fast_vindex=fast_vindex,
        )



import logging

logger = logging.getLogger("pytcube")

LOG_FORMAT = (
    "%(asctime)s|%(levelname)-8s|%(message)-100s|%(filename)s:%(lineno)d@%(funcName)s()"
)


def setup_logging(
    app_logger: logging.Logger,
) -> bool:
    """Set up logging for the specified logger."""

    app_logger.setLevel(logging.DEBUG)
    app_logger.handlers = []
    handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt=LOG_FORMAT)
    handler.setFormatter(formatter)
    app_logger.addHandler(handler)

    return True



from pytcube.testing.grid import generate_grid, generate_spatio_temporal_grid
from pytcube.testing.point import (generate_point,
                                   generate_spatio_temporal_point)

__all__ = [
    "generate_grid",
    "generate_spatio_temporal_grid",
    "generate_point",
    "generate_spatio_temporal_point",
]



"""
Voici les tests à effectuer:
* [x] Test basic
* [x] Test multiple_points
* [x] Test use obs
* [x] Test search_windows
* [x] Test cyclic_dims
* [x] Test prefix
* [x] Test other dimensions
* [x] Test without fast_vindex
* [x] Test sur N dimensions
* [x] Test Indexation sur 2 dimensions sur 3
"""

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from pytcube.testing import generate_grid


def test_colocation_basic():
    """Test that colocation returns a dataset with expected shape."""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
    result = grid.pytcube.colocation(dim0=3.1, dim1=6.9)
    values = result["var0"].values
    expected = np.array([[["(3, 7)"]]], dtype=object)

    assert isinstance(result, xr.Dataset)
    # Should only return one point
    assert result.sizes["dim0_step"] == 1
    assert result.sizes["dim1_step"] == 1
    assert np.array_equal(values, expected)


def test_colocation_multiple_points():
    """Test multiple points selection."""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
    result = grid.pytcube.colocation(
        dim0=[3.1, 7.2],
        dim1=[6.9, 5.5],
    )
    values = result["var0"].values
    expected = np.array([[["(3, 7)"]], [["(7, 6)"]]], dtype=object)

    assert np.array_equal(values, expected)


def test_colocation_use_obs():
    """Test colocation with obs arg."""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
    result = grid.pytcube.colocation(
        obs=[39, 57],
        dim0=[3.1, 7.2],
        dim1=[6.9, 5.5],
    )
    assert 39 in result.obs.values
    assert 57 in result.obs.values


def test_colocation_search_window():
    """Test search_windows"""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
    result = grid.pytcube.colocation(
        dim0=[3.1, 7.2], dim1=[6.9, 5.5], search_window={"dim0": 1, "dim1": 1}
    )
    values = result["var0"].isel(obs=0).values
    expected = np.array(
        [
            ["(2, 6)", "(2, 7)", "(2, 8)"],
            ["(3, 6)", "(3, 7)", "(3, 8)"],
            ["(4, 6)", "(4, 7)", "(4, 8)"],
        ],
        dtype=object,
    )

    assert np.array_equal(values, expected)


def test_colocation_cyclic_dims():
    """Test cyclic_dims"""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
    result = grid.pytcube.colocation(
        dim0=[0.1, 7.2],
        dim1=[6.9, 0.3],
        search_window={"dim0": 1, "dim1": 1},
        cyclic_dims=["dim0"],
    )

    assert 0 in result.obs.values  # cyclic_dims a fonctionné
    assert (
        1 not in result.obs.values
    )  # l'observation est en dehors du domaine, elle n'est donc pas selectionné.


def test_colocation_prefix():
    """Test point & grid prefix"""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
    result = grid.pytcube.colocation(
        dim0=[3.1, 7.2],
        dim1=[6.9, 5.5],
        grid_prefix="GRD_",
        point_prefix="PNT_",
    )

    assert "GRD_dim0" in result.coords
    assert "PNT_dim0" in result.coords


def test_colocation_other_dimensions():
    """Test colocation with other dimension in grid."""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="dindices", chunks=5)
    grid["other_dim"] = ("other_dim", np.arange(20))
    grid["var1"] = ("other_dim", da.random.random(20, chunks=10))

    result = grid.pytcube.colocation(
        dim0=[3.1, 7.2], dim1=[6.9, 5.5], search_window={"dim0": 1, "dim1": 1}
    )
    result = result.compute()
    assert "grid_other_dim" in result.coords
    assert "var1" in result.data_vars


def test_colocation_without_fast_vindex():
    """Test colocation without fast_vindex."""

    grid = generate_grid(nvar=1, shape=(10, 10), fmt="dindices", chunks=5)
    result = grid.pytcube.colocation(
        dim0=[3.1, 7.2],
        dim1=[6.9, 5.5],
        search_window={"dim0": 1, "dim1": 1},
        fast_vindex=False,
    )
    result = result.compute()
    values = result["var0"].isel(obs=0).values
    expected = np.array(
        [
            ["(2, 6)", "(2, 7)", "(2, 8)"],
            ["(3, 6)", "(3, 7)", "(3, 8)"],
            ["(4, 6)", "(4, 7)", "(4, 8)"],
        ],
        dtype=object,
    )
    assert np.array_equal(values, expected)


class TestColocationNDimensions:
    """Test la colocation sur plusieurs dimension"""

    def test_1D(self):
        grid = generate_grid(nvar=1, shape=(10,), fmt="nindices")
        result = grid.pytcube.colocation(
            dim0=[3.1, 7.2],
            search_window={"dim0": 1},
        )
        assert isinstance(result, xr.Dataset)
        assert True  # juste pour expliciter que le test vérifie l’absence d’erreur

    def test_2D(self):
        grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")
        result = grid.pytcube.colocation(
            dim0=[3.1, 7.2],
            dim1=[6.9, 5.5],
            search_window={"dim0": 1, "dim1": 1},
        )
        assert isinstance(result, xr.Dataset)
        assert True

    def test_3D(self):
        grid = generate_grid(nvar=1, shape=(10, 10, 10), fmt="nindices")
        result = grid.pytcube.colocation(
            dim0=[3.1, 7.2],
            dim1=[6.9, 5.5],
            dim2=[4.6, 2.9],
            search_window={"dim0": 1, "dim1": 1, "dim2": 1},
        )
        assert isinstance(result, xr.Dataset)
        assert True

    def test_4D(self):
        grid = generate_grid(nvar=1, shape=(10, 10, 10, 10), fmt="nindices")
        result = grid.pytcube.colocation(
            dim0=[3.1, 7.2],
            dim1=[6.9, 5.5],
            dim2=[4.6, 2.9],
            dim3=[8.3, 3.5],
            search_window={"dim0": 1, "dim1": 1, "dim2": 1, "dim3": 1},
        )
        assert isinstance(result, xr.Dataset)
        assert True


class TestColocationNotAllDimension:
    """Test la colocation sur des dimensions manquantes"""

    def test_with_narray(self):
        """Test avec des variable de type NumPy array"""
        grid = generate_grid(nvar=1, shape=(10, 10), fmt="nindices")

        result = grid.pytcube.colocation(
            dim0=[3.1, 7.2],
            search_window={"dim0": 1},
        )

        assert "grid_dim1" in result.coords

    def test_with_darray(self):
        """Test avec des variable de type Dask array
        Erreur du à Fast-Vindex, nécéssité de faire évoluer la librairie
        """
        grid = generate_grid(nvar=1, shape=(10, 10), fmt="dindices", chunks=5)

        with pytest.raises(ValueError):
            result = grid.pytcube.colocation(
                dim0=[3.1, 7.2],
                search_window={"dim0": 1},
            )
            assert isinstance(result, xr.Dataset)



import numpy as np
import pytest
import xarray as xr

from pytcube.core.point import set_points


def test_set_points_with_lists():
    ds = set_points(obs=[0, 1, 2], lat=[10, 20, 30], lon=[1, 2, 3])
    assert isinstance(ds, xr.Dataset)
    assert np.array_equal(ds.coords["obs"].values, [0, 1, 2])
    assert np.array_equal(ds["lat"].values, [10, 20, 30])
    assert np.array_equal(ds["lon"].values, [1, 2, 3])


def test_set_points_with_numpy_arrays():
    obs = np.array([0, 1])
    lat = np.array([45.0, 46.0])
    lon = np.array([-73.0, -74.0])
    ds = set_points(obs=obs, lat=lat, lon=lon)
    assert isinstance(ds, xr.Dataset)
    assert np.array_equal(ds.coords["obs"].values, obs)
    assert np.array_equal(ds["lat"].values, lat)
    assert np.array_equal(ds["lon"].values, lon)


def test_set_points_with_scalars():
    ds = set_points(obs=0, lat=10, lon=20)
    assert isinstance(ds, xr.Dataset)
    # obs et data_vars doivent être convertis en array de longueur 1
    assert ds.sizes["obs"] == 1
    assert np.array_equal(ds.coords["obs"].values, [0])
    assert np.array_equal(ds["lat"].values, [10])
    assert np.array_equal(ds["lon"].values, [20])


def test_set_points_with_xarray():
    obs = xr.DataArray([0, 1, 2])
    lat = xr.DataArray([10, 20, 30])
    ds = set_points(obs=obs, lat=lat)
    assert isinstance(ds, xr.Dataset)
    assert np.array_equal(ds.coords["obs"].values, [0, 1, 2])
    assert np.array_equal(ds["lat"].values, [10, 20, 30])


def test_set_points_without_obs():
    ds = set_points(lat=[1, 2, 3], lon=[4, 5, 6])
    # obs doit être généré automatiquement
    assert np.array_equal(ds.coords["obs"].values, [0, 1, 2])
    assert np.array_equal(ds["lat"].values, [1, 2, 3])
    assert np.array_equal(ds["lon"].values, [4, 5, 6])


def test_set_points_inconsistent_lengths():
    with pytest.raises(ValueError):
        set_points(obs=[0, 1], lat=[10, 20, 30], lon=[1, 2])



import datetime

import numpy as np
import pandas as pd
import xarray as xr


def _narray_datetime(t_min, t_max, size):
    """Generate a NumPy array of random datetime values between t_min and t_max."""
    total_seconds = (t_max - t_min).total_seconds()
    return np.array(
        [
            t_min + datetime.timedelta(seconds=np.random.randint(0, int(total_seconds)))
            for _ in range(size)
        ]
    )


def generate_spatio_temporal_point(
    nobs: int, grid: xr.Dataset, padding: int = 0
) -> xr.Dataset:
    """Generate a spatio-temporal point Dataset with random coordinates within the grid limits."""

    obs = np.arange(nobs)
    coords = {"obs": ("obs", obs)}

    # For each dimension in the grid, generate random values within its limits
    for dim in ["time", "lon", "lat"]:
        d_min = np.sort(grid[dim].values)[padding]
        d_max = np.sort(grid[dim].values)[-(padding + 1)]
        if dim == "time":
            # Convert min and max to datetime
            t_min = pd.to_datetime(d_min)
            t_max = pd.to_datetime(d_max)
            coords["time"] = ("obs", _narray_datetime(t_min, t_max, nobs))
        else:
            coords[dim] = ("obs", np.random.uniform(d_min, d_max, nobs))

    return xr.Dataset(coords=coords)


def generate_point(nobs: int, grid: xr.Dataset, padding: int = 0) -> xr.Dataset:
    """Generate a point Dataset with random coordinates for each grid dimension."""

    obs = np.arange(nobs)
    coords = {"obs": ("obs", obs)}

    # For each grid dimension, generate random coordinates within its limits
    for dim in grid.dims:
        d_min = np.sort(grid[dim].values)[padding]
        d_max = np.sort(grid[dim].values)[-(padding + 1)]
        coords[dim] = ("obs", np.random.uniform(d_min, d_max, nobs))

    return xr.Dataset(coords=coords)



import numpy as np

from pytcube.utils.core import find_nearest_indices


def test_find_nearest_indices():
    """Test the function to find the nearest indices in a coordinate array for given observation points."""

    coord_array = np.arange(-179, 181, 2)
    obs_array = np.array([23.12, 46.89, -78.62, -121.97, -179.97, 179.97])

    idx = find_nearest_indices(coord_array, obs_array)
    idx_expected = np.array([101, 113, 50, 29, 0, 179])

    assert np.array_equal(idx, idx_expected)



import numpy as np
import xarray as xr

from pytcube.utils.types import ensure_numpy_array


def test_ensure_numpy_array_from_numpy():
    arr = np.array([1, 2, 3])
    result = ensure_numpy_array(arr)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, arr)


def test_ensure_numpy_array_from_list():
    data = [1, 2, 3]
    result = ensure_numpy_array(data)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, np.array(data))


def test_ensure_numpy_array_from_scalar():
    data = 42
    result = ensure_numpy_array(data)
    assert isinstance(result, np.ndarray)
    assert result.shape == (1,)
    assert result[0] == 42


def test_ensure_numpy_array_from_xarray():
    da = xr.DataArray([1, 2, 3])
    result = ensure_numpy_array(da)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, da.values)


def test_ensure_numpy_array_from_nested_list():
    data = [[1, 2], [3, 4]]
    result = ensure_numpy_array(data)
    assert isinstance(result, np.ndarray)
    assert result.shape == (2, 2)
    assert np.array_equal(result, np.array(data))


def test_ensure_numpy_array_from_empty_list():
    data = []
    result = ensure_numpy_array(data)
    assert isinstance(result, np.ndarray)
    assert result.size == 0



from pytcube.core.accessor import PytCubeAccessor
from pytcube.core.grid import GridAccessor
from pytcube.core.point import PointAccessor

__version__ = "0.0.0"
__all__ = [
    PytCubeAccessor,
    GridAccessor,
    PointAccessor,
]



import xarray as xr


@xr.register_dataset_accessor("grid")
class GridAccessor:
    def __init__(self, xarray_obj: xr.Dataset):
        self._obj = xarray_obj

    def _format(self, prefix):
        dims_renamed = {dim: f"{prefix}{dim}" for dim in self._obj.dims}
        return self._obj.rename(dims_renamed)  # rename dims and coords
